#include <MD_Potential.h>
#include <Print.h>

#define DEF_TM

using namespace ZSET;

double MD_POTENTIAL::get_potential(ARRAY<PARTICLE*> p)
{
  if (name == LJ)
  {
     if (p.size() != 2)
       ERROR("<LJ> potential takes only two points");
     // For similar materials
     double n = param[0],
   	    m = param[1],
	    rij = (p[1]->x - p[0]->x).norm(),
	    eps, sig, s_r;
     if (initialized)
     { 
	      eps = coef[0];
	      sig = coef[1]; 
     }
     // Different materials
     else
     {
	ARRAY<double> mix_mat;
        mix_mat = mixing_rule(p[0]->mat,p[1]->mat);
	eps = mix_mat[0]; 
	sig = mix_mat[1];
     }
     s_r = sig/rij;
     //prn("Potential mat <"+itoa(p[0]->mat->id)+">, <"+itoa(p[1]->mat->id)+"> :: eps = "+formatd(eps)+", sigma = ", sig);
     
     return alpha*eps*(pow(s_r,n)-pow(s_r,m));
    // Add to different materials
  }
  else if (name == TM)
  {
     if (p.size() != 2)
       ERROR("<Two minima> potential takes only two points");
     // For similar materials
     double n = param[0],
 	    m = param[1],
            c = param[2],
            d = param[3],
	    eps = coef[0],
	    sig = coef[1],
	    rij = (p[1]->x - p[0]->x).norm(),
	    s_r = sig/rij,
	    r2 = sqrt(2.)*pow(2.,1./6.)*sig;
     
     return alpha*eps*( (pow(s_r,n)-pow(s_r,m)) + c*exp(-d*pow((rij-r2)/sig,2.)) );
     //return alpha*eps*( (pow(s_r,n)-pow(s_r,m)) + c*(1.-sin((rij-r2)/sig))*exp(-d*pow((rij-r2)/sig,2.)) );
    // Add to different materials
  }
  else if (name == CONTACT)
  {
     if (p.size() != 2)
       ERROR("<CONTACT> potential takes only two points");
     return 0;
     // For similar materials
     double sigma_max = coef[0],
            dissip    = coef[1],
            rij = (p[1]->x - p[0]->x).norm();

    // Add to different materials
  }
  else if (name == EAM)
  {
	//bla-bla-bla // To Michael
  }
  else
    ERROR("Unknown potential");
}
double MD_POTENTIAL::get_potential(PARTICLE* p1, PARTICLE* p2)
{
  ARRAY<PARTICLE*> p(2);
  p[0] = p1;
  p[1] = p2;
  return get_potential(p);
}
VECTOR MD_POTENTIAL::get_force(ARRAY<PARTICLE*> p)
{
  if (name == LJ)
  {
     if (p.size() != 2)
       ERROR("LJ potential takes only two points");
     double n = param[0],
            m = param[1];
	    
     if (p[0]->mat->id == p[1]->mat->id) // Same material
     {
       if (initialized)
       {
         double eps = coef[0],
                sig = coef[1],
	        rij = (p[1]->x - p[0]->x).norm(),
	        s_r = sig/rij;
         return alpha*eps/(rij*rij)*(n*pow(s_r,n)-m*pow(s_r,m))*(p[1]->x - p[0]->x);
       }
       else
       {
	 double eps = p[0]->mat->material_parameter[0],
                sig = p[0]->mat->material_parameter[1],
                rij = (p[1]->x - p[0]->x).norm(),
                s_r = sig/rij;

         return alpha*eps/(rij*rij)*(n*pow(s_r,n)-m*pow(s_r,m))*(p[1]->x - p[0]->x);
       }
     }
     else // Different materials
     {
         ARRAY<double> eps_sig_mix = mixing_rule(p[0]->mat,p[1]->mat);
	 double eps = eps_sig_mix[0],
                sig = eps_sig_mix[1],
                rij = (p[1]->x - p[0]->x).norm(),
                s_r = sig/rij;
         //prn("Force mat <"+itoa(p[0]->mat->id)+">, <"+itoa(p[1]->mat->id)+"> :: eps = "+formatd(eps)+", sigma = ", sig);

         return alpha*eps/(rij*rij)*(n*pow(s_r,n)-m*pow(s_r,m))*(p[1]->x - p[0]->x);
     }
  }
  else if (name == CONTACT)
  {
     // material_parameter[0], R, Particle radius 
     // material_parameter[1], E, Young's modulus
     // material_parameter[2], v, Poisson's ratio
     // param[0] - distance for the max force
     // param[1] - separation distance 
     if (p.size() != 2)
       ERROR("<CONTACT> potential takes only two points");
     // For similar materials
     double dmax = param[0],
            dsep = param[1],
            rij = (p[1]->x - p[0]->x).norm(),
	    dc = 0.5*(p[0]->mat->material_parameter[0] + p[1]->mat->material_parameter[0]),
	    R  = 1./(1./p[0]->mat->material_parameter[0] + 1./p[1]->mat->material_parameter[0]),
	    E  = 1./( (1.-p[0]->mat->material_parameter[2]*p[0]->mat->material_parameter[2])/p[0]->mat->material_parameter[1] + (1.-p[1]->mat->material_parameter[2]*p[1]->mat->material_parameter[2])/p[1]->mat->material_parameter[1]),
	    En = E/(R*R),
	    D  = damage[ p[0]->id ][ p[1]->id ],
	    xi = dmax/(dsep-dmax);
     if (D > 0. && rij > dc) { // elasticity and fracture
       double dmax_prime = dmax*(1 + xi)/(D + xi);
       if ( (rij-dc) < dmax_prime )
	 return En*D*(rij-dc);
       else {
	 damage[ p[0]->id ][ p[1]->id ] = damage[ p[1]->id ][ p[0]->id ] = dmax * (1 - (rij-dc)*xi + xi*dmax )/(rij-dc);
	 return En*D*dmax_prime + En*(dmax - (rij - dc)*xi);
       }
     }
     else { // pure contact 
       return -1.333333*sqrt(R)*E*pow(dc-rij,1.5);
     }
    // Add to different materials
  }
  else if (name == TM)
  {
     if (p.size() != 2)
       ERROR("LJ potential takes only two points");
     // For similar materials
     double n = param[0],
 	    m = param[1],
            c = param[2],
            d = param[3],
	    eps = coef[0],
	    sig = coef[1],
	    rij = (p[1]->x - p[0]->x).norm(),
	    s_r = sig/rij,
	    r2 = sqrt(2.)*pow(2.,1./6.)*sig;
     
     return alpha*eps/(rij*rij)*(n*pow(s_r,n)-m*pow(s_r,m) + c*rij/sig*exp(-d*pow((rij-r2)/sig,2.))*2*d*(rij-r2)/sig )*(p[1]->x - p[0]->x); 
//     return alpha*eps/(rij*rij)*(n*pow(s_r,n)-m*pow(s_r,m) + c*rij/sig*exp(-d*pow((rij-r2)/sig,2.))*(cos((rij-r2)/sig)+(1-sin((rij-r2)/sig))*2*d*(rij-r2)/sig ))*(p[1]->x - p[0]->x); 
  }
  else
    ERROR("Unknown potential");
}
VECTOR MD_POTENTIAL::get_force(PARTICLE* p1, PARTICLE* p2)
{
  ap[0] = p1;
  ap[1] = p2;

  return get_force(ap);
}
void MD_POTENTIAL::set_mix_rule(ARRAY<double> parameters)
{
  if_mix_rule = TRUE;
//  mix.resize(parameters.size());
  mix = parameters;
}
ARRAY<double> MD_POTENTIAL::mixing_rule(MD_MATERIAL* mat1, MD_MATERIAL* mat2)
{
  if (mat1->name == mat2->name && mat2->name == LJ) {
    ARRAY<double> result(2);
    if (mat1->id == mat2->id)
    {
      result[0] = mat1->material_parameter[0];
      result[1] = mat1->material_parameter[1];
    }
    else if (if_mix_rule)
    {
      if ((int(mix[0]) == mat1->id && int(mix[1]) == mat2->id) || (int(mix[0]) == mat2->id && int(mix[1]) == mat1->id))
      {
        result[0] = mix[2];
        result[1] = mix[3];
      }
    } 
    else
    {
      result[0] = sqrt(mat1->material_parameter[0]*mat2->material_parameter[0]);
      result[1] = 0.5*(mat1->material_parameter[1]+mat2->material_parameter[1]);
    }
    return result;
  }
  else
    ERROR("Do not know mixing rules for not LJ potentials");
}
void MD_POTENTIAL::initialize(MD_MATERIAL* mat)
{
  if (name == LJ || name == TM)
  {
    alpha = 1./(param[0]+param[1])*pow(pow(param[0],param[0])/pow(param[1],param[1]),1./(param[0]-param[1]));
    coef.resize(2);
    coef[0] = mat->material_parameter[0];
    coef[1] = mat->material_parameter[1];
    initialized = TRUE;
  }
  else if (name == CONTACT) {
    coef.resize(3);
    coef[0] = mat->material_parameter[0];
    coef[1] = mat->material_parameter[1];
    coef[2] = mat->material_parameter[2];
  } 
  else
    ERROR("Unknown potential");
}
void MD_POTENTIAL::initialize()
{
  if (name == LJ || name == TM)
  {
    alpha = 1./(param[0]+param[1])*pow(pow(param[0],param[0])/pow(param[1],param[1]),1./(param[0]-param[1]));
    initialized = FALSE;
  }
  else
    ERROR("Unknown potential");
}
void MD_POTENTIAL::scale_parameters(MULTIPLIERS m)
{
   if (name == LJ)
   {
     if (mix.size() == 4)
     {
       mix[2] /= m.energy;
       mix[3] /= m.length;
     }
     else
	ERROR("Mix rule has not been assigned");
   }
   else
     ERROR("Unknown potential <"+name+">");
}
void MD_POTENTIAL::precompute_links(ARRAY<CELL*>& cells, int psize) {
  
  damage.resize(psize);
  for (int i = 0; i < damage.size(); i++) {
    damage[i].resize(psize);
    for (int j = 0; j < damage[i].size(); j++) 
      damage[i][j] = 0;
  }

  double dist, dsep;
  LIST<int> neighbours; 
      for (int icell = 0; icell < cells.size(); icell++)
      {
	neighbours = cells[icell]->get_neighbours();
        cells[icell]->checked = TRUE;
        for (int ip = 0; ip < cells[icell]->particles.size(); ip++)
        {
	  PARTICLE& pi = *(cells[icell]->particles[ip]);
// Check in the current cell
            for (int jp = ip+1; jp < cells[icell]->particles.size(); jp++)
            {
	      PARTICLE& pj = *(cells[icell]->particles[jp]);
	      dist = (pi.x-pj.x).norm();	
	      dsep = param[1];
	      if (dist < dsep)
		damage[ pi.id ][ pj.id] = damage[ pj.id ][ pi.id] = 1.;
	    }
// Check in neighbouring cells
//    Middle cells 
	    for (int in = 0; in < neighbours.size(); in++)
            {
	      if (!cells[neighbours[in]]->checked)
              {
		for (int jp = 0; jp < cells[neighbours[in]]->particles.size(); jp++)
                {
             	  PARTICLE& pj = *(cells[neighbours[in]]->particles[jp]);
       	          dist = (pi.x-pj.x).norm();	
	          dsep = coef[1];
	          if (dist < dsep)
		    damage[ pi.id ][ pj.id] = damage[ pj.id ][ pi.id] = 1.;
          	}
	      }
            }
	}
      }
      for (int icell = 0; icell < cells.size(); icell++)
        cells[icell]->checked = FALSE; 
}

